<?php

namespace Laravelista\Localized;

use Illuminate\Http\Request;
use Illuminate\Routing\Route;
use Illuminate\Routing\Router;
use Illuminate\Routing\RouteUrlGenerator;
use Illuminate\Routing\UrlGenerator;
use Illuminate\Support\Collection;
use Illuminate\Support\Facades\App;
use Illuminate\Support\Facades\Config;

class Localized
{
    protected RouteUrlGenerator $routeUrlGenerator;

    public function __construct(
        protected Router $router,
        protected Request $request,
        protected UrlGenerator $urlGenerator
    ) {
        $this->routeUrlGenerator = new RouteUrlGenerator($this->urlGenerator, $this->request);
    }

    protected function getDefaultLocale(): string
    {
        /**
         * @var string
         */
        return Config::get('localized.defaultLocale');
    }

    protected function getAppLocale(): string
    {
        return App::getLocale();
    }

    protected function allRoutes(): array
    {
        return $this->router->getRoutes()->getRoutes();
    }

    protected function routes(): Collection
    {
        return collect($this->allRoutes());
    }

    protected function createRequest(
        string $path,
        array $parameters = []
    ): Request {
        $tail = implode('/', array_map(
            fn (string $parameter) => rawurlencode($parameter),
            $this->urlGenerator->formatParameters($parameters)
        ));

        $uri = trim($path.'/'.$tail, '/');

        return new Request(server: ['REQUEST_URI' => $uri]);
    }

    public function getParameters(
        array $localizedParameters,
        Route $currentRoute
    ): array {
        if (count($localizedParameters) > 0) {
            return $localizedParameters;
        }
        
        $currentRouteParameters = $currentRoute->parameters();

        // This handles the situation where Route::view is used.
        if (array_key_exists('data', $currentRouteParameters)) {
            return $currentRouteParameters['data'];
        }

        return $currentRouteParameters;
    }

    public function getParametersForLocale(
        array $parameters,
        string $locale
    ): array {
        if (array_key_exists($locale, $parameters)) {
            /**
             * @var array
             */
            return $parameters[$locale];
        }

        return [];
    }

    public function getCurrentRoute(): ?Route
    {
        /**
         * @var Route|null $currentRoute
         */
        $currentRoute = $this->request->route();

        if (is_null($currentRoute)) {
            return null;
        }

        if (!array_key_exists('identifier', $currentRoute->action)) {
            return null;
        }

        return $currentRoute;
    }

    public function getLocalizedCurrentRoute(
        Route $currentRoute,
        string $locale,
        string $defaultLocale
    ): ?Route {
        /**
         * @var Route|null
         */
        return $this->routes()
            ->reject(fn (Route $route) =>
                (!array_key_exists('identifier', $route->action)))
            ->filter(fn (Route $route) =>
                ($route->action['identifier'] === $currentRoute->action['identifier'] and
                 ((!array_key_exists('locale', $route->action) and
                   $locale === $defaultLocale)
                  or
                  (array_key_exists('locale', $route->action) and
                   $route->action['locale'] === $locale))))
            ->first();
    }

    /**
     * Returns the route name for given route name and locale.
     *
     * This method is meant to be used with the `route()` helper and is the method that
     * you will probably use the most.
     *
     * ### Usage
     *
     * If the default locale set is `en`, for routes defined in `routes/web.php` as:
     *
     * ```php
     * Route::get('/', [Controller::class, 'home'])->name('home')->identifier('home');
     * Route::get('/hr', [Controller::class, 'home'])->name('home')->identifier('home')->locale('hr');
     *
     * Route::get('articles/{article}, [Controller::class, 'show'])->name('articles.show')->identifier('articles.show');
     *
     * Route::get('/hr/clanci/{article}, [Controller::class, 'show'])->name('hr.articles.show')->identifier('articles.show');
     * ```
     *
     * When you need to create a link to a named route use this method like so:
     *
     * ```php
     * <a href="{{ route(Localized::route('home')) }}">Home</a>
     *
     * // This is the same as `route('home')`.
     *
     * // Or
     *
     * <a href="{{ route(Localized::route('articles.show', 'hr'), 'the-article-slug') }}">Home</a>
     *
     * // Which is the same as `route('hr.articles.show', 'the-article-slug')`.
     * ```
     */
    public function route(
        string $name,
        ?string $locale = null
    ): string {
        $wantedRoute = $this->router->getRoutes()->getByName($name);

        if (is_null($wantedRoute)) {
            return $name;
        }

        $targetLocale = $locale ?? $this->getAppLocale();

        if ($targetLocale === $this->getDefaultLocale()) {
            return $name;
        }

        if (!array_key_exists('identifier', $wantedRoute->action)) {
            return $name;
        }

        /**
         * @var Route|null $localizedWantedRoute
         */
        $localizedWantedRoute = $this->routes()
            ->reject(fn (Route $route) =>
                (!array_key_exists('identifier', $route->action) or
                 !array_key_exists('locale', $route->action)))
            ->filter(fn (Route $route) =>
                ($route->action['identifier'] === $wantedRoute->action['identifier'] and
                 $route->action['locale'] === $targetLocale))
            ->first();

        if (is_null($localizedWantedRoute)) {
            return $name;
        }

        if (!array_key_exists('as', $localizedWantedRoute->action)) {
            return $name;
        }

        /**
         * @var string
         */
        return $localizedWantedRoute->action['as'];
    }

    /**
     * Generates a URL based on given path and locale.
     *
     * The parameters for this method are almost the same as parameters for
     * `UrlGenerator::to` method, except that this method accepts an additional
     * parameter called `locale`.
     *
     * Based on the given path it tries to get a localized "wanted" route. If this route
     * is not found it fallbacks to the `UrlGenerator::to` method. Otherwise, it
     * glues parameters to the found route Uri.
     *
     * ### Usage
     *
     * For routes defined in `routes/web.php` as:
     *
     * ```php
     * Route::get('manufacturer/{manufacturer}/serial-number/{serialNumber}, [Controller::class, 'show'])->identifier('manufacturer.serial-number');
     *
     * Route::get('hr/proizvodac/{manufacturer}/serijski-broj/{serialNumber}, [Controller::class, 'show'])->identifier('manufacturer.serial-number')->locale('hr');
     * ```
     *
     * Calling this method as:
     *
     * ```php
     * Localized::url('/manufacturer', ['bmw', 'serial-number', '1111-1111'], locale: 'hr')
     * ```
     *
     * would result in the following URL:
     *
     * ```
     * /hr/proizvodac/bmw/serijski-broj/1111-1111
     * ```
     *
     * @todo The example above is not correct. This method needs more work.
     * It would return `/hr/proizvodac/bmw/serial-number/1111-1111`.
     */
    public function url(
        string $path,
        array $parameters = [],
        ?bool $secure = null,
        ?string $locale = null
    ): string {
        $routes = $this->routes();

        /**
         * @var Route|null $wantedRoute
         */
        $wantedRoute = $routes
            ->filter(fn (Route $route) =>
                $route->matches($this->createRequest($path, $parameters)))
            ->first();

        if (is_null($wantedRoute)) {
            return $this->urlGenerator->to($path, $parameters, $secure);
        }

        if (!array_key_exists('identifier', $wantedRoute->action)) {
            return $this->urlGenerator->to($path, $parameters, $secure);
        }

        $defaultLocale = $this->getDefaultLocale();

        $targetLocale = $locale ?? $this->getAppLocale();

        /**
         * @var Route|null $localizedWantedRoute
         */
        $localizedWantedRoute = $routes
            ->reject(fn (Route $route) => !array_key_exists('identifier', $route->action))
            ->filter(function (Route $route) use ($wantedRoute, $targetLocale, $defaultLocale) {
                if ($targetLocale === $defaultLocale) {
                    return $route->action['identifier'] === $wantedRoute->action['identifier'] and
                        (!array_key_exists('locale', $route->action) or
                         $route->action['locale'] === $defaultLocale);
                }

                if (!array_key_exists('locale', $route->action)) {
                    return false;
                }

                return $route->action['identifier'] === $wantedRoute->action['identifier'] and
                    $route->action['locale'] === $targetLocale;
            })
            ->first();

        if (is_null($localizedWantedRoute)) {
            return $this->urlGenerator->to($path, $parameters, $secure);
        }

        $firstPart = rtrim(explode('{', $localizedWantedRoute->uri)[0], '/');

        return $this->urlGenerator->to($firstPart, $parameters, $secure);
    }

    /**
     * Generates a URL for the current route based on the given locale and parameters.
     *
     * It tries to generate a URL for the current route based on given locale and parameters.
     * If it is unable to generate a URL, it returns the URL to the `fallback` parameter.
     *
     * ### Parameters
     *
     * If the parameters are the same for each locale (they are not localized), then the parameters
     * are an array or key/value based pairs. eg:
     *
     * ```php
     * $parameters = [
     *     'slug' => 'house'
     * ];
     * ```
     *
     * If the parameters are different for each locale, then the user needs to provide parameters
     * grouped by the locale. eg:
     *
     * ```php
     * $parameters = [
     *     'en' => [
     *         'slug' => 'house'
     *     ],
     *     'hr' => [
     *         'slug' => 'kuća'
     *     ]
     * ];
     * ```
     *
     * ### Usage
     *
     * This method is usually called when building a language picker component.
     *
     * Somewhere in your layout view, eg. `layouts/base.blade.php`:
     *
     * ```php
     * <ul>
     *     @foreach(config('localized.locales') as $locale)
     *         <li>
     *             <a href="{{ Localized::get($locale, $localizedParameters ?? []) }}">
     *                 {{ $locale }}
     *             </a>
     *         </li>
     *     @endforeach
     * </ul>
     * ```
     *
     * In your view, eg. `show.blade.php`:
     *
     * ```php
     * @extends('layouts.base', [
     *     'localizedParameters' => [
     *         'slug' => 'house'
     *     ]
     * ])
     * ```
     *
     * - `$localizedParameters` - This variable name can be any.
     * - `config('localized.locales')` - Locales for this loop can be used from anywhere.
     */
    public function get(
        string $locale,
        array $parameters = [],
        string $fallback = '/'
    ): string {
        $currentRoute = $this->getCurrentRoute();

        if (is_null($currentRoute)) {
            return $this->urlGenerator->to($fallback);
        }

        $localizedCurrentRoute = $this->getLocalizedCurrentRoute(
            $currentRoute,
            $locale,
            $this->getDefaultLocale()
        );

        if (is_null($localizedCurrentRoute)) {
            return $this->urlGenerator->to($fallback);
        }

        $localizedParameters = $this->getParametersForLocale($parameters, $locale);

        return $this->routeUrlGenerator->to(
            $localizedCurrentRoute,
            $this->getParameters($localizedParameters, $currentRoute)
        );
    }
}
