<?php

namespace Laravelista\Localized\Tests;

use Illuminate\Http\Request;
use Illuminate\Routing\Route;
use Illuminate\Routing\RouteCollection;
use Illuminate\Routing\Router;
use Illuminate\Routing\RouteUrlGenerator;
use Illuminate\Routing\UrlGenerator;
use Illuminate\Support\Facades\App;
use Illuminate\Support\Facades\Config;
use Laravelista\Localized\Localized;
use Mockery;
use Mockery\MockInterface;
use PHPUnit\Framework\TestCase;

class LocalizedTest extends TestCase
{
    public function tearDown(): void
    {
        Mockery::close();
    }

    protected function getMockedRouter(): MockInterface
    {
        /**
         * @var MockInterface
         */
        $m = Mockery::mock('alias:' . Router::class);

        return $m;
    }

    protected function getMockedRequest(): MockInterface
    {
        /**
         * @var MockInterface
         */
        $m = Mockery::mock(Request::class);

        return $m;
    }

    protected function getMockedUrlGenerator(): MockInterface
    {
        /**
         * @var MockInterface
         */
        $m = Mockery::mock(UrlGenerator::class);

        return $m;
    }

    protected function getMockedRouteUrlGenerator(): MockInterface
    {
        /**
         * @var MockInterface
         */
        $m = Mockery::mock(RouteUrlGenerator::class);

        return $m;
    }

    /**
     * This test ensures that class Localized can be instantiated.
     */
    public function testCanBeInstantiated()
    {
        $router = $this->getMockedRouter();

        $request = $this->getMockedRequest();

        $urlGenerator = $this->getMockedUrlGenerator();

        $routeUrlGenerator = $this->getMockedRouteUrlGenerator();

        /**
         * @var Router $router
         * @var Request $request
         * @var UrlGenerator $urlGenerator
         * @var RouteUrlGenerator $routeUrlGenerator
         */
        $localized = new Localized($router, $request, $urlGenerator, $routeUrlGenerator);

        $this->assertInstanceOf(Localized::class, $localized);
    }

    /**
     * When given route name and desired locale, the
     * method should return the name of the route for given locale.
     */
    public function testCanLocalizeRouteNameBasedOnGivenLocale()
    {
        // getDefaultLocale
        Config::shouldReceive('get')
            ->once()
            ->with('localized.defaultLocale')
            ->andReturn('en');

        $homeEnRoute = new Route('GET', '/', [
            'as' => 'home',
            'identifier' => 'home'
        ]);

        $homeHrRoute = new Route('GET', '/hr', [
            'as' => 'hr.home',
            'identifier' => 'home',
            'locale' => 'hr'
        ]);

        $routeCollection = new RouteCollection();
        $routeCollection->add($homeEnRoute);
        $routeCollection->add($homeHrRoute);

        $router = $this->getMockedRouter();
        $router->shouldReceive('getRoutes')
            ->times(2)
            ->andReturn($routeCollection);

        $request = $this->getMockedRequest();

        $urlGenerator = $this->getMockedUrlGenerator();

        $routeUrlGenerator = $this->getMockedRouteUrlGenerator();

        /**
         * @var Router $router
         * @var Request $request
         * @var UrlGenerator $urlGenerator
         * @var RouteUrlGenerator $routeUrlGenerator
         */
        $localized = new Localized($router, $request, $urlGenerator, $routeUrlGenerator);

        $this->assertEquals('hr.home', $localized->route('home', 'hr'));
    }

    /**
     * When given route name and default locale, the
     * method should return the name of the route for default locale.
     */
    public function testCanLocalizeRouteNameBasedOnGivenDefaultLocale()
    {
        // getDefaultLocale
        Config::shouldReceive('get')
            ->once()
            ->with('localized.defaultLocale')
            ->andReturn('en');

        $homeEnRoute = new Route('GET', '/', [
            'as' => 'home',
            'identifier' => 'home'
        ]);

        $homeHrRoute = new Route('GET', '/hr', [
            'as' => 'hr.home',
            'identifier' => 'home',
            'locale' => 'hr'
        ]);

        $routeCollection = new RouteCollection();
        $routeCollection->add($homeEnRoute);
        $routeCollection->add($homeHrRoute);

        $router = $this->getMockedRouter();
        $router->shouldReceive('getRoutes')
            ->times(1)
            ->andReturn($routeCollection);

        $request = $this->getMockedRequest();

        $urlGenerator = $this->getMockedUrlGenerator();

        $routeUrlGenerator = $this->getMockedRouteUrlGenerator();

        /**
         * @var Router $router
         * @var Request $request
         * @var UrlGenerator $urlGenerator
         * @var RouteUrlGenerator $routeUrlGenerator
         */
        $localized = new Localized($router, $request, $urlGenerator, $routeUrlGenerator);

        $this->assertEquals('home', $localized->route('home', 'en'));
    }

    /**
     * This test should return the name of the route for the current
     * application locale. This is the most used scenario.
     */
    public function testCanLocalizeRouteNameBasedOnAppLocale()
    {
        // getDefaultLocale
        Config::shouldReceive('get')
            ->once()
            ->with('localized.defaultLocale')
            ->andReturn('en');

        // getAppLocale
        App::shouldReceive('getLocale')
            ->once()
            ->andReturn('hr');

        $homeEnRoute = new Route('GET', '/', [
            'as' => 'home',
            'identifier' => 'home',
            'locale' => 'en'
        ]);

        $homeHrRoute = new Route('GET', '/hr', [
            'as' => 'hr.home',
            'identifier' => 'home',
            'locale' => 'hr'
        ]);

        $routeCollection = new RouteCollection();
        $routeCollection->add($homeEnRoute);
        $routeCollection->add($homeHrRoute);

        $router = $this->getMockedRouter();
        $router->shouldReceive('getRoutes')
            ->times(2)
            ->andReturn($routeCollection);

        $request = $this->getMockedRequest();

        $urlGenerator = $this->getMockedUrlGenerator();

        $routeUrlGenerator = $this->getMockedRouteUrlGenerator();

        /**
         * @var Router $router
         * @var Request $request
         * @var UrlGenerator $urlGenerator
         * @var RouteUrlGenerator $routeUrlGenerator
         */
        $localized = new Localized($router, $request, $urlGenerator, $routeUrlGenerator);

        $this->assertEquals('hr.home', $localized->route('home'));
    }

    /**
     * The test tries to get a localized version of a route which
     * does not have a localized version for given locale.
     *
     * It should return the passed in name as fallback.
     */
    public function testCannotLocalizeRouteNameBasedOnGivenLocale()
    {
        // getDefaultLocale
        Config::shouldReceive('get')
            ->once()
            ->with('localized.defaultLocale')
            ->andReturn('en');

        $homeEnRoute = new Route('GET', '/', [
            'as' => 'home',
            'identifier' => 'home'
        ]);

        $routeCollection = new RouteCollection();
        $routeCollection->add($homeEnRoute);

        $router = $this->getMockedRouter();
        $router->shouldReceive('getRoutes')
            ->times(2)
            ->andReturn($routeCollection);

        $request = $this->getMockedRequest();

        $urlGenerator = $this->getMockedUrlGenerator();

        $routeUrlGenerator = $this->getMockedRouteUrlGenerator();

        /**
         * @var Router $router
         * @var Request $request
         * @var UrlGenerator $urlGenerator
         * @var RouteUrlGenerator $routeUrlGenerator
         */
        $localized = new Localized($router, $request, $urlGenerator, $routeUrlGenerator);

        $this->assertEquals('home', $localized->route('home', 'hr'));
    }

    /**
     * This test ensures that the method for obtaining
     * the localizedCurrentRoute is valid.
     */
    public function testCanGetLocalizedCurrentRoute()
    {
        $aboutEnRoute = new Route('GET', '/about', [
            'as' => 'about',
            'identifier' => 'about'
        ]);

        $aboutHrRoute = new Route('GET', '/hr/o-nama', [
            'as' => 'hr.about',
            'identifier' => 'about',
            'locale' => 'hr'
        ]);

        $routeCollection = new RouteCollection();
        $routeCollection->add($aboutEnRoute);
        $routeCollection->add($aboutHrRoute);

        $router = $this->getMockedRouter();
        $router->shouldReceive('getRoutes')
            ->times(3)
            ->andReturn($routeCollection);

        $request = $this->getMockedRequest();

        $urlGenerator = $this->getMockedUrlGenerator();

        $routeUrlGenerator = $this->getMockedRouteUrlGenerator();

        /**
         * @var Router $router
         * @var Request $request
         * @var UrlGenerator $urlGenerator
         * @var RouteUrlGenerator $routeUrlGenerator
         */
        $localized = new Localized($router, $request, $urlGenerator, $routeUrlGenerator);

        $localizedCurrentRoute = $localized->getLocalizedCurrentRoute($aboutEnRoute, 'hr', 'en');

        $this->assertEquals('hr.about', $localizedCurrentRoute?->getAction('as'));

        $localizedCurrentRoute = $localized->getLocalizedCurrentRoute($aboutEnRoute, 'en', 'en');

        $this->assertEquals('about', $localizedCurrentRoute?->getAction('as'));

        $localizedCurrentRoute = $localized->getLocalizedCurrentRoute($aboutHrRoute, 'en', 'en');

        $this->assertEquals('about', $localizedCurrentRoute?->getAction('as'));
    }

    /**
     * This test ensures that the method for obtaining
     * the localizedCurrentRoute is valid.
     */
    public function testCannotGetLocalizedCurrentRoute()
    {
        $aboutEnRoute = new Route('GET', '/about', [
            'as' => 'about',
            'identifier' => 'about'
        ]);

        $routeCollection = new RouteCollection();
        $routeCollection->add($aboutEnRoute);

        $router = $this->getMockedRouter();
        $router->shouldReceive('getRoutes')
            ->times(1)
            ->andReturn($routeCollection);

        $request = $this->getMockedRequest();

        $urlGenerator = $this->getMockedUrlGenerator();

        $routeUrlGenerator = $this->getMockedRouteUrlGenerator();

        /**
         * @var Router $router
         * @var Request $request
         * @var UrlGenerator $urlGenerator
         * @var RouteUrlGenerator $routeUrlGenerator
         */
        $localized = new Localized($router, $request, $urlGenerator, $routeUrlGenerator);

        $localizedCurrentRoute = $localized->getLocalizedCurrentRoute($aboutEnRoute, 'hr', 'en');

        $this->assertNull($localizedCurrentRoute?->getAction('as'));
    }

    /**
     * This test ensures that the method for obtaining
     * the currentRoute is valid.
     */
    public function testCanGetCurrentRoute()
    {
        $aboutEnRoute = new Route('GET', '/about', [
            'as' => 'about',
            'identifier' => 'about'
        ]);

        $router = $this->getMockedRouter();

        $request = $this->getMockedRequest();
        $request->shouldReceive('route')
            ->times(1)
            ->andReturn($aboutEnRoute);

        $urlGenerator = $this->getMockedUrlGenerator();

        $routeUrlGenerator = $this->getMockedRouteUrlGenerator();

        /**
         * @var Router $router
         * @var Request $request
         * @var UrlGenerator $urlGenerator
         * @var RouteUrlGenerator $routeUrlGenerator
         */
        $localized = new Localized($router, $request, $urlGenerator, $routeUrlGenerator);

        $currentRoute = $localized->getCurrentRoute();

        $this->assertEquals('about', $currentRoute?->getAction('as'));
    }

    /**
     * This test ensures that the method for obtaining
     * the currentRoute is valid.
     *
     * Even if the currentRoute is found, if it does not have an
     * identifier set it returns null.
     */
    public function testCannotGetCurrentRouteWhenIdentifierIsMissing()
    {
        $aboutEnRoute = new Route('GET', '/about', [
            'as' => 'about',
        ]);

        $router = $this->getMockedRouter();

        $request = $this->getMockedRequest();
        $request->shouldReceive('route')
            ->times(1)
            ->andReturn($aboutEnRoute);

        $urlGenerator = $this->getMockedUrlGenerator();

        $routeUrlGenerator = $this->getMockedRouteUrlGenerator();

        /**
         * @var Router $router
         * @var Request $request
         * @var UrlGenerator $urlGenerator
         * @var RouteUrlGenerator $routeUrlGenerator
         */
        $localized = new Localized($router, $request, $urlGenerator, $routeUrlGenerator);

        $currentRoute = $localized->getCurrentRoute();

        $this->assertNull($currentRoute?->getAction('as'));
    }

    /**
     * It tests that the method getParametersForLocale
     * return localized parameters if any.
     */
    public function testCanGetParametersForLocale()
    {
        $router = $this->getMockedRouter();

        $request = $this->getMockedRequest();

        $urlGenerator = $this->getMockedUrlGenerator();

        $routeUrlGenerator = $this->getMockedRouteUrlGenerator();

        /**
         * @var Router $router
         * @var Request $request
         * @var UrlGenerator $urlGenerator
         * @var RouteUrlGenerator $routeUrlGenerator
         */
        $localized = new Localized($router, $request, $urlGenerator, $routeUrlGenerator);

        $parameters = [
            'slug' => 'house'
        ];

        $this->assertEquals([], $localized->getParametersForLocale($parameters, 'hr'));

        $parameters = [
            'hr' => [
                'slug' => 'house'
            ]
        ];

        $this->assertEquals($parameters['hr'], $localized->getParametersForLocale($parameters, 'hr'));
    }

    /**
     * It tests that the method getParameters
     * return localized parameters if any otherwise it
     * returns the route parameters.
     */
    public function testCanGetParameters()
    {
        $router = $this->getMockedRouter();

        $request = $this->getMockedRequest();

        $urlGenerator = $this->getMockedUrlGenerator();

        $routeUrlGenerator = $this->getMockedRouteUrlGenerator();

        /**
         * @var Router $router
         * @var Request $request
         * @var UrlGenerator $urlGenerator
         * @var RouteUrlGenerator $routeUrlGenerator
         */
        $localized = new Localized($router, $request, $urlGenerator, $routeUrlGenerator);

        $currentRoute = new Route('GET', '/about', [
            'as' => 'about',
        ]);

        $parameters = [
            'slug' => 'house'
        ];

        $this->assertEquals($parameters, $localized->getParameters($parameters, $currentRoute));

        $currentRoute = new Route('GET', '/about', [
            'as' => 'about'
        ]);
        $currentRoute->parameters = $parameters = [
            'slug' => 'apartment'
        ];

        $this->assertEquals($parameters, $localized->getParameters([], $currentRoute));
    }
}
